use Test::More tests => 1;

BEGIN {
use_ok( 'DBIx::Class::ResultSet::RecursiveUpdate' );
}

diag( "Testing DBIx::Class::ResultSet::RecursiveUpdate $DBIx::Class::ResultSet::RecursiveUpdate::VERSION" );
